/*
  Alarm Detection

  Detects the repeating pattern of fire alarms via a microphone, and alerts the user via vibration motor.

  The circuit:
  MAX4466 Microphone on analog input 1
  On-board LED on digital output 1
  Vibraion motor with transistor on analog output 4

  Created 24/6/2017
  By Cooper Lawrence
  Last Modified 16/7/2016
  By Cooper Lawrence
*/

const int recordedArrayLength = 100; //Length of the recorded signals
const int signalSampleLength = 50; //Length of time to sample the microphone's output

const int sampleThreshold = 700; //What constitutes a "loud" noise
const int convolutionThreshold = 30; //The maximum difference between the two signals to count as a match

boolean prevRecordedSound[recordedArrayLength]; //Last recorded signal
boolean recordedSound[recordedArrayLength]; //Current recorded signal

int prevConvolutionResult = 0; //Last result of the convolution between the two signals
int convolutionResult = 0; //Current result of the convolution between the two signals

boolean alarmDetected = false;

void setup() {
  pinMode(1, OUTPUT); //Allow the board's LED to be used
}

void loop() {
  if (alarmDetected == true) {
    vibrate(1);
  } else {
    sample(recordedSound, signalSampleLength); //Record sound from the microphone

    //Check if there's an alarm going off
    if (checkForAlarm() == true) {
      alarmDetected = true;
    } else if (checkForLoudNoise() == true) { //Check if there's a loud sustained noise
      vibrate(3);
    }

    //Copy the current signal into the previous signal
    for (int i = 0; i < recordedArrayLength; i++) {
      prevRecordedSound[i] = recordedSound[i];
    }

    //Copy the current convolution result into the previous convolution result
    prevConvolutionResult = convolutionResult;
  }
}

//Vibration patterns for the user to choose from
void vibrate(int patternNumber) {
  switch (patternNumber) {
    case 0:
      //Add code to vibrate with the same pattern as the alarm
      break;
    case 1:
      analogWrite(4, 255);
      delay(1000);
      analogWrite(4, 0);
      break;

    case 2:
      analogWrite(4, 255);
      delay(1000);
      analogWrite(4, 0);
      delay(1000);
      break;

    case 3:
      for (int i = 0; i < 2; i++) {
        analogWrite(4, 255);
        delay(200);
        analogWrite(4, 0);
        delay(200);
      }
      analogWrite(4, 255);
      delay(500);
      analogWrite(4, 0);
      delay(300);
      break;
  }
}

//Check if there's a loud sustained noise
boolean checkForLoudNoise() {
  if (convolution(recordedSound, recordedSound) > 90) {
    return true;
  }

  return false;
}

//Check if there's an alarm going off
boolean checkForAlarm() {
  convolutionResult = convolution(recordedSound, prevRecordedSound); //Perform convolution between the two signals

  int signalDifference = abs(convolutionResult - prevConvolutionResult); //Calculate the difference between the two signals

  //Check if the two signals are a match (the difference between the two is less than the set threshold, and they aren't silent)
  if (signalDifference < convolutionThreshold && convolutionResult > 10 && prevConvolutionResult > 10) {
    return true;
  }

  return false;
}

//Convolution algorithm to match the two signals
int convolution(boolean currentArray[], boolean lastArray[]) {
  int maxSum = 0;

  for (int i = 0; i < recordedArrayLength; i++) {
    int tmpSum = 0;

    for (int h = i; h >= 0; h--) {
      int k = i - h;

      tmpSum += currentArray[k] * lastArray[(recordedArrayLength - 1) - h];
    }

    if (tmpSum > maxSum) {
      maxSum = tmpSum;
    }
  }

  return maxSum;
}

void sample(boolean array[], int sampleLength) {
  for (int i = 0; i < recordedArrayLength; i++) {
    long startMillis = millis(); //Start of sample window
    int peakToPeak = 0; //Peak-to-peak level

    int signalMax = 0;
    int signalMin = 1024;

    //Collect data for the specified length of time
    while (millis() - startMillis < sampleLength) {
      int sample = analogRead(1);

      if (sample < 1024) {  //Toss out spurious readings
        if (sample > signalMax) {
          signalMax = sample;  //Save just the max levels
        } else if (sample < signalMin) {
          signalMin = sample;  //Save just the min levels
        }
      }
    }

    peakToPeak = signalMax - signalMin;  //Peak to peak amplitude

    if (peakToPeak >= sampleThreshold) {
      recordedSound[i] = 1;

      digitalWrite(1, HIGH);
    } else {
      recordedSound[i] = 0;

      digitalWrite(1, LOW);
    }
  }
}
